import java.util.Iterator;

public class HashTable<E extends KeyedElementInterface<K>, K> implements HashTableInterface<E, K> {

	public static enum ProbeType {Linear, Quadtratic, DoubleHashing}
	
	private static final int NUMBER_OF_BUCKETS = 17_321;
	
	private Bucket[] buckets;
	private int size;
	private ProbeType probeType = ProbeType.Linear;
	
	
	
	private int numProbes;
	
	
	public HashTable() {
		this.buckets = new Bucket[NUMBER_OF_BUCKETS];
		for ( int i = 0 ; i < this.buckets.length ; i++ ) {
			this.buckets[i] = new Bucket();
			this.buckets[i].element = null;
			this.buckets[i].emptyFromBeginning = true;
		}
		this.size = 0;
		this.numProbes = 0;
	}
	
	@Override
	public Iterator<E> iterator() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public int size() {
		return this.size;
	}

	@Override
	public boolean isEmpty() {
		return (this.size == 0);
	}

	@Override
	public HashTableInterface<E, K> copy() throws InstantiationException {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public E insert(E element) {
		K key = element.getKey();
		int hashCode = key.hashCode() % this.buckets.length;

		int probes = 1;
		if (this.size == this.buckets.length) {
			while (probes < this.buckets.length) {
				// check if we need to replace
				if (((E) this.buckets[hashCode].element).getKey().equals(key)) {
					E oldElement = (E) this.buckets[hashCode].element;
					this.buckets[hashCode].element = element;
					return oldElement;
				}
				
				hashCode = this.probe(hashCode, probes++);
			}
			this.numProbes += probes;
		}
		
		while (this.buckets[hashCode].element != null) {
			// check if we need to replace
			if (((E) this.buckets[hashCode].element).getKey().equals(key)) {
				E oldElement = (E) this.buckets[hashCode].element;
				this.buckets[hashCode].element = element;
				return oldElement;
			}
			
			hashCode = this.probe(hashCode, probes++);
		}
		
		this.buckets[hashCode].element = element;
		this.buckets[hashCode].emptyFromBeginning = false;
		
		this.size++;
		this.numProbes += probes;
		
		return null;
	}

	@Override
	public E get(K key) {
		int hashCode = key.hashCode() % this.buckets.length; 
		
		int probes = 1;
		while ((!this.buckets[hashCode].emptyFromBeginning) && (probes < this.buckets.length)) {
			if (((E) this.buckets[hashCode].element).getKey().equals(key)) {
				this.numProbes += probes;
				return (E) this.buckets[hashCode].element;
			}
			
			hashCode = this.probe(hashCode, probes++);
		}
		
		this.numProbes += probes;
		return null;
	}

	@Override
	public E remove(K key) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void removeAll() {
		this.buckets = new Bucket[NUMBER_OF_BUCKETS];
		for ( int i = 0 ; i < this.buckets.length ; i++ ) {
			this.buckets[i] = new Bucket();
			this.buckets[i].element = null;
			this.buckets[i].emptyFromBeginning = true;
		}
		this.size = 0;
		this.numProbes = 0;
	}

	@Override
	public int getSizeOfLargestBucket() {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public double getAverageBucketSize() {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public Object[] getBuckets() {
		// TODO Auto-generated method stub
		return null;
	}
	
	public int getNumProbes() {
		return numProbes;
	}

	public void setNumProbes(int numProbes) {
		this.numProbes = numProbes;
	}

	public ProbeType getProbeType() {
		return probeType;
	}

	public void setProbeType(ProbeType probeType) {
		this.probeType = probeType;
	}

	/*******************/
	/* private methods */
	/*******************/

	private int probe(int hashCode, int i) {
		switch (probeType) {
		case Linear:
			return (hashCode + 1) % this.buckets.length;
		case Quadtratic:
			return (hashCode + i + (i * i)) % this.buckets.length;
		case DoubleHashing:
			return  (hashCode + i * (int)( 7919.0/953.0) * hashCode) % this.buckets.length;
		default:
			return 0;
		}
	}
	
}


class Bucket<E> {
	public boolean emptyFromBeginning;
	public E element;
}
